import numpy as np
import deepxde as dde
from ..utils import Math
from ..cases.PDECases import Burgers
from ..solver.PDESolver import PINNSolver


class FI_PINN(PINNSolver):
    def __init__(
        self,
        PDECase,
        max_iter=10,
        epsilon_r=0.1,
        epsilon_p=0.1,
        max_iter_SAIS=100,
        N_1=300,
        N_2=1000,
        p0=0.1,
    ):
        self.max_iter = max_iter
        self.epsilon_r = epsilon_r
        self.epsilon_p = epsilon_p
        self.max_iter_SAIS = max_iter_SAIS
        self.N_1 = N_1
        self.N_2 = N_2
        self.p0 = p0
        super().__init__(name="FI-PINN", PDECase=PDECase)

    def g(self, X):
        Y = np.abs(self.model.predict(X, operator=self.PDECase.pde)).astype(np.float64)
        if Y.shape[0] != X.shape[0]:
            Y = np.sum(Y, axis=0)
        return Y - self.epsilon_r

    def random_points(self, N, mul, sigma, sample_rate=10):
        result = []
        while True:
            X = self.PDECase.geomtime.random_points(N * sample_rate)
            X = Math.transform_uniform_to_normal_2D(X)
            X = Math.transform_normal_to_truncated_normal_on_geomtime(
                X, self.PDECase.geomtime, mul, sigma
            )
            if X.shape[0] >= N:
                result = X[0:N, :]
                break
            else:
                result.append(X)
                N -= X.shape[0]
        return np.array(result)

    def get_probility(self, X, loc, scale):
        from scipy import stats

        y = stats.norm.pdf(X, loc=loc, scale=scale)
        return y

    def SAIS(self):
        mul_i = np.array([0.0, 0.0])
        sigma_i = np.array([1.0])
        for i in range(self.max_iter_SAIS):
            h = self.random_points(self.N_1, mul_i, sigma_i)
            h_g = self.g(h).squeeze()
            index = np.argsort(-h_g)
            h_g = h_g[index]
            h = h[index, :]
            N_eta = h_g[0]
            N_p = int(np.floor(self.p0 * self.N_1))
            h = h[0:N_p, :]
            if N_eta <= N_p:
                h_p = self.get_probility(h, np.array([0.0, 0.0]), np.array([1.0, 1.0]))
                mul_i = np.sum(h * h_p, axis=0) / np.sum(h_p)
                sigma_i = np.sum((h - mul_i) ** 2, axis=0) / (N_p - 1)
            else:
                break
        mul_opt = np.mean(h[0:N_p, :], axis=0)
        sigma_opt = np.var(h[0:N_p, :], axis=0, ddof=1)

        print(f"mul_opt:{mul_opt}, sigma_opt:{sigma_opt}")

        N2_sample = self.random_points(self.N_2, mul_opt, sigma_opt)

        Pf = (
            np.sum(
                (self.g(N2_sample) > 0)
                * self.get_probility(
                    N2_sample, np.array([0.0, 0.0]), np.array([1.0, 1.0])
                )
                / self.get_probility(N2_sample, mul_i, sigma_i)
            )
            / self.N_2
        )

        N2_g = self.g(N2_sample).squeeze()
        # if np.mean(N2_g) < self.epsilon_r:
        # self.epsilon_r = np.mean(N2_g)
        index = np.where(N2_g > 0)
        N2_g = N2_g[index]
        D_adaptive = N2_sample[index]

        return Pf, D_adaptive

    def closure(self):
        self.train_step()

        for i in range(self.max_iter):
            Pf, D_adaptive = self.SAIS()
            print(f"iter:{i}, Pf:{Pf}")
            if Pf < self.epsilon_p:
                break
            if D_adaptive.shape[0] != 0:
                print(f"iter:{i}, add {D_adaptive.shape[0]} samples")
                self.PDECase.data.add_anchors(D_adaptive)
                self.PDECase.Visualization.plot_data_2D(D_adaptive)
            else:
                print("D_adaptive.shape[0] = 0:")
            self.train_step(iterations=1000, lr=0.0001)


if __name__ == "__main__":
    PDECase = Burgers(NumDomain=2000)
    solver = FI_PINN(PDECase=PDECase)
    solver.train()
    solver.save(add_time=True)



__all__ = [
    "FuncCases",
    "FuncFromFormula",
    "FuncFromData",
    "PoissonCases",
    "PoissonCase1D",
    "PoissonCase2D",
    "Poisson_1D_Dirichlet",
    "Poisson_1D_Dirichlet_Neumann",
    "Poisson_1D_Dirichlet_Robin",
    "Poisson_1D_Dirichlet_Periodic",
    "Poisson_1D_Dirichlet_PointSetOperator",
    "Poisson_1D_Hard_Boundary",
    "Poisson_1D_Fourier_Net",
    "Poisson_2D_L_Shaped",
    "Poisson_1D_Unknown_Forcing_Field_Inverse",
    "Poisson_1D_Fractional_Inverse",
    "Poisson_2D_Fractional_Inverse",
    "Poisson_2D_Peak",
    "PDECases",
    "A_Simple_ODE",
    "LotkaVolterra",
    "SecondOrderODE",
    "Laplace_disk",
    "Helmholtz",
    "Helmholtz_Hole",
    "Helmholtz_Sound_hard_Absorbing",
    "Kovasznay_Flow",
    "Euler_Beam",
    "Burgers",
    "Heat",
    "Diffusion",
    "Diffusion_reaction",
    "AllenCahn",
    "Klein_Gordon",
    "Beltrami_flow",
    "WaveCase1D",
    "Wave_1D_STMsFFN",
    "Wave_1D_Hard_Boundary",
    "Schrodinger",
    "IDE",
    "Volterra_IDE",
    "Fractional_Poisson_1D",
    "Fractional_Poisson_2D",
    "Fractional_Poisson_3D",
    "Fractional_Diffusion_1D",
    "InverseCase",
    "Lorenz_Inverse",
    "Lorenz_Exogenous_Input_Inverse",
    "Brinkman_Forchheimer_Inverse",
    "Diffusion_Inverse",
    "Diffusion_Reaction_Inverse",
    "Navier_Stokes_Incompressible_Flow_Around_Cylinder_Inverse",
    "Bimodal_2D",
    "NS_Flow_in_LidDriven_Cavity",
]

from .PDECases import (
    Burgers,
    Heat,
    AllenCahn,
    Diffusion,
    Klein_Gordon,
    Beltrami_flow,
    A_Simple_ODE,
    LotkaVolterra,
    SecondOrderODE,
    Laplace_disk,
    Helmholtz,
    Helmholtz_Hole,
    Helmholtz_Sound_hard_Absorbing,
    Kovasznay_Flow,
    Euler_Beam,
    Diffusion_reaction,
    IDE,
    Volterra_IDE,
    Fractional_Poisson_1D,
    Fractional_Poisson_2D,
    Fractional_Poisson_3D,
    Fractional_Diffusion_1D,
    Schrodinger,
    Bimodal_2D,
    NS_Flow_in_LidDriven_Cavity,
)
from .FuncCases import FuncCases, FuncFromFormula, FuncFromData
from .PoissonCases import (
    PoissonCase1D,
    PoissonCase2D,
    Poisson_1D_Dirichlet,
    Poisson_1D_Dirichlet_Neumann,
    Poisson_1D_Dirichlet_Robin,
    Poisson_1D_Dirichlet_Periodic,
    Poisson_1D_Dirichlet_PointSetOperator,
    Poisson_1D_Hard_Boundary,
    Poisson_1D_Fourier_Net,
    Poisson_2D_L_Shaped,
    Poisson_1D_Unknown_Forcing_Field_Inverse,
    Poisson_1D_Fractional_Inverse,
    Poisson_2D_Fractional_Inverse,
    Poisson_2D_Peak,
)
from .WaveCases import WaveCase1D, Wave_1D_STMsFFN, Wave_1D_Hard_Boundary
from .InverseCases import (
    InverseCase,
    Lorenz_Inverse,
    Lorenz_Exogenous_Input_Inverse,
    Brinkman_Forchheimer_Inverse,
    Diffusion_Inverse,
    Diffusion_Reaction_Inverse,
    Navier_Stokes_Incompressible_Flow_Around_Cylinder_Inverse,
)



import deepxde as dde
from ..cases.PDECases import Diffusion
from ..solver.PDESolver import PINNSolver


class Random_R(PINNSolver):
    def __init__(self, PDECase, P):
        super().__init__(name=f"Random_R_P_{P}", PDECase=PDECase)
        self.P = P
        self.resampler = dde.callbacks.PDEPointResampler(period=self.P)

    def closure(self):
        self.train_step(callbacks=[self.resampler])


if __name__ == "__main__":
    PDECase = Diffusion(NumDomain=2000)
    solver = Random_R(P=00, PDECase=PDECase)
    solver.train()
    solver.save(add_time=True)



import numpy as np
import matplotlib.pyplot as plt


class Visualization_1D:
    def __init__(self, x_label=None, y_label=None):
        self.x_label = x_label
        self.y_label = y_label

    def set_axes_1D(self, axes=None, title=None):
        kwargs = [self.x_label, self.y_label, title]
        if axes is None:
            fig, axes = plt.subplots()
        fun = [axes.set_xlabel, axes.set_ylabel, axes.set_title]
        for i, f in enumerate(fun):
            if kwargs[i] is not None:
                f(kwargs[i])
        return axes

    def plot_line_1D(self, PDECase, solver, exact=True, **kwargs):
        axes = self.set_axes_1D(**kwargs, title=PDECase.name)

        X, y = PDECase.get_testdata()
        if exact:
            axes.plot(X, y, label="Exact")
        axes.plot(X, solver.model.predict(X), "--", label="Prediction")
        axes.legend()
        return axes


class Visualization_2D:
    def __init__(
        self,
        x_limit=None,
        y_limit=None,
        x_label=None,
        y_label=None,
        feature_transform=None,
    ):
        self.x_limit = x_limit
        self.y_limit = y_limit
        self.x_label = x_label
        self.y_label = y_label
        self.feature_transform = feature_transform

    def set_axes_2D(self, axes=None, title=None):
        kwargs = [self.x_limit, self.y_limit, self.x_label, self.y_label, title]
        if axes is None:
            fig, axes = plt.subplots()
        fun = [
            axes.set_xlim,
            axes.set_ylim,
            axes.set_xlabel,
            axes.set_ylabel,
            axes.set_title,
        ]
        for i, f in enumerate(fun):
            if kwargs[i] is not None:
                f(kwargs[i])
        return axes

    def plot_data_2D(self, X, **kwargs):
        axes = self.set_axes_2D(**kwargs)
        if self.feature_transform is not None:
            X = self.feature_transform(X)
        axes.scatter(X[:, 0], X[:, 1])
        plt.show()
        return axes

    def plot_heatmap_2D(self, X, y, shape, **kwargs):
        axes = self.set_axes_2D(**kwargs)
        if self.feature_transform is not None:
            X = self.feature_transform(X)
        return axes.pcolormesh(
            X[:, 0].reshape(shape[0], shape[1]),
            X[:, 1].reshape(shape[0], shape[1]),
            y.reshape(shape[0], shape[1]),
            cmap="rainbow",
        )

    def plot_exact_predict_error_2D(
        self, x, y_true, y_pred, shape, title, colorbar=[True, True, True]
    ):
        fig, axes = plt.subplots(1, 3, figsize=(15, 5))
        axs = []
        axs.append(
            self.plot_heatmap_2D(
                x, y_true, shape=shape, axes=axes[0], title="Exact solution"
            )
        )
        axs.append(
            self.plot_heatmap_2D(x, y_pred, shape=shape, axes=axes[1], title=title)
        )
        axs.append(
            self.plot_heatmap_2D(
                x,
                np.abs(y_pred - y_true),
                shape=shape,
                axes=axes[2],
                title="Absolute error",
            )
        )

        for needColorbar, ax, axe in zip(colorbar, axs, axes):
            if needColorbar:
                fig.colorbar(ax, ax=axe)
        plt.show()
        return fig, axes



import numpy as np
import deepxde as dde
import deepxde.backend as bkd
from abc import abstractmethod
from .PDECases import PDECases
from ..utils import *


class InverseCase(PDECases):
    def __init__(
        self,
        name,
        external_trainable_variables,
        NumDomain=2000,
        use_output_transform=False,
        layer_size=[2] + [32] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
        metrics=None,
        loss_weights=None,
        visualization=None,
    ):
        super().__init__(
            name=name,
            NumDomain=NumDomain,
            use_output_transform=use_output_transform,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            external_trainable_variables=external_trainable_variables,
            metrics=metrics,
            loss_weights=loss_weights,
            visualization=visualization,
        )

    @abstractmethod
    def gen_pde(self):
        pass

    @abstractmethod
    def gen_geomtime(self):
        pass

    @abstractmethod
    def gen_data(self):
        pass


class Lorenz_Inverse(InverseCase):
    def __init__(
        self,
        NumDomain=400,
        layer_size=[1] + [40] * 3 + [3],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.C1 = dde.Variable(1.0)
        self.C2 = dde.Variable(1.0)
        self.C3 = dde.Variable(1.0)
        super().__init__(
            "Inverse problem for the Lorenz system",
            external_trainable_variables=[self.C1, self.C2, self.C3],
            NumDomain=NumDomain,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            visualization=Visualization_1D(x_label="t", y_label="y"),
        )

    def gen_testdata(self):
        import os

        basepath = os.path.abspath(__file__)
        folder = os.path.dirname(os.path.dirname(basepath))
        data_path = os.path.join(folder, "data/Lorenz.npz")
        data = np.load(data_path)
        return data["t"], data["y"]

    def gen_pde(self):
        def Lorenz_system(x, y):
            y1, y2, y3 = y[:, 0:1], y[:, 1:2], y[:, 2:]
            dy1_x = dde.grad.jacobian(y, x, i=0)
            dy2_x = dde.grad.jacobian(y, x, i=1)
            dy3_x = dde.grad.jacobian(y, x, i=2)
            return [
                dy1_x - self.C1 * (y2 - y1),
                dy2_x - y1 * (self.C2 - y3) + y2,
                dy3_x - y1 * y2 + self.C3 * y3,
            ]

        return Lorenz_system

    def gen_geomtime(self):
        return dde.geometry.TimeDomain(0, 3)

    def gen_data(self):
        ic1 = dde.icbc.IC(
            self.geomtime, lambda X: -8, lambda _, on_initial: on_initial, component=0
        )
        ic2 = dde.icbc.IC(
            self.geomtime, lambda X: 7, lambda _, on_initial: on_initial, component=1
        )
        ic3 = dde.icbc.IC(
            self.geomtime, lambda X: 27, lambda _, on_initial: on_initial, component=2
        )
        observe_t, ob_y = self.get_testdata()
        observe_y0 = dde.icbc.PointSetBC(observe_t, ob_y[:, 0:1], component=0)
        observe_y1 = dde.icbc.PointSetBC(observe_t, ob_y[:, 1:2], component=1)
        observe_y2 = dde.icbc.PointSetBC(observe_t, ob_y[:, 2:3], component=2)
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            [ic1, ic2, ic3, observe_y0, observe_y1, observe_y2],
            num_domain=self.NumDomain,
            num_boundary=2,
            anchors=observe_t,
        )

    def plot_result(self, solver, axes=None, exact=True):
        C1_true = 10
        C2_true = 15
        C3_true = 8 / 3

        C1_pred = bkd.to_numpy(self.C1)
        C2_pred = bkd.to_numpy(self.C2)
        C3_pred = bkd.to_numpy(self.C3)

        C1_error = np.abs(C1_true - C1_pred)
        C2_error = np.abs(C2_true - C2_pred)
        C3_error = np.abs(C3_true - C3_pred)

        print(f"C1 true: {C1_true}, C2 true: {C2_true}, C3 true: {C3_true}")
        print(f"C1 pred: {C1_pred}, C2 pred: {C2_pred}, C3 pred: {C3_pred}")
        print(f"C1 error: {C1_error}, C2 error: {C2_error}, C3 error: {C3_error}")

        axes = self.Visualization.plot_line_1D(self, solver, exact, axes=axes)
        return axes


class Lorenz_Exogenous_Input_Inverse(InverseCase):
    def __init__(
        self,
        NumDomain=400,
        layer_size=[1] + [40] * 3 + [3],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.C1 = dde.Variable(1.0)
        self.C2 = dde.Variable(1.0)
        self.C3 = dde.Variable(1.0)
        self.C1_true = 10
        self.C2_true = 15
        self.C3_true = 8 / 3
        super().__init__(
            "Inverse problem for the Lorenz system with exogenous input",
            external_trainable_variables=[self.C1, self.C2, self.C3],
            NumDomain=NumDomain,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            visualization=Visualization_1D(x_label="t", y_label="y"),
        )

    def gen_testdata(self):
        import scipy as sp
        from scipy.integrate import odeint

        def ex_func(t):
            spline = sp.interpolate.Rbf(
                time, ex_input, function="thin_plate", smooth=0, episilon=0
            )
            return spline(t)

        def LorezODE(x, t):
            x1, x2, x3 = x
            dxdt = [
                self.C1_true * (x2 - x1),
                x1 * (self.C2_true - x3) - x2,
                x1 * x2 - self.C3_true * x3 + ex_func(t),
            ]
            return dxdt

        maxtime = 3
        time = np.linspace(0, maxtime, 200)
        ex_input = 10 * np.sin(2 * np.pi * time)
        x0 = [-8, 7, 27]
        x = odeint(LorezODE, x0, time)
        return time[:, np.newaxis], x

    def gen_pde(self):
        def Lorenz_system(x, y, ex):
            y1, y2, y3 = y[:, 0:1], y[:, 1:2], y[:, 2:]
            dy1_x = dde.grad.jacobian(y, x, i=0)
            dy2_x = dde.grad.jacobian(y, x, i=1)
            dy3_x = dde.grad.jacobian(y, x, i=2)
            ex = bkd.from_numpy(ex)
            return [
                dy1_x - self.C1 * (y2 - y1),
                dy2_x - y1 * (self.C2 - y3) + y2,
                dy3_x - y1 * y2 + self.C3 * y3 - ex,
            ]

        return Lorenz_system

    def gen_geomtime(self):
        return dde.geometry.TimeDomain(0, 3)

    def gen_data(self):
        ic1 = dde.icbc.IC(
            self.geomtime, lambda X: -8, lambda _, on_initial: on_initial, component=0
        )
        ic2 = dde.icbc.IC(
            self.geomtime, lambda X: 7, lambda _, on_initial: on_initial, component=1
        )
        ic3 = dde.icbc.IC(
            self.geomtime, lambda X: 27, lambda _, on_initial: on_initial, component=2
        )
        observe_t, ob_y = self.get_testdata()
        observe_y0 = dde.icbc.PointSetBC(observe_t, ob_y[:, 0:1], component=0)
        observe_y1 = dde.icbc.PointSetBC(observe_t, ob_y[:, 1:2], component=1)
        observe_y2 = dde.icbc.PointSetBC(observe_t, ob_y[:, 2:3], component=2)

        def ex_func2(t):
            import scipy as sp

            maxtime = 3
            time = np.linspace(0, maxtime, 200)
            ex_input = 10 * np.sin(2 * np.pi * time)
            spline = sp.interpolate.Rbf(
                time, ex_input, function="thin_plate", smooth=0, episilon=0
            )
            return spline(t[:, 0:])

        return dde.data.PDE(
            self.geomtime,
            self.pde,
            [ic1, ic2, ic3, observe_y0, observe_y1, observe_y2],
            num_domain=self.NumDomain,
            num_boundary=2,
            anchors=observe_t,
            auxiliary_var_function=ex_func2,
        )

    def plot_result(self, solver, axes=None, exact=True):
        C1_pred = bkd.to_numpy(self.C1)
        C2_pred = bkd.to_numpy(self.C2)
        C3_pred = bkd.to_numpy(self.C3)

        C1_error = np.abs(self.C1_true - C1_pred)
        C2_error = np.abs(self.C2_true - C2_pred)
        C3_error = np.abs(self.C3_true - C3_pred)

        print(
            f"C1 true: {self.C1_true}, C2 true: {self.C2_true}, C3 true: {self.C3_true}"
        )
        print(f"C1 pred: {C1_pred}, C2 pred: {C2_pred}, C3 pred: {C3_pred}")
        print(f"C1 error: {C1_error}, C2 error: {C2_error}, C3 error: {C3_error}")

        axes = self.Visualization.plot_line_1D(solver, exact, axes=axes)
        return axes


class Brinkman_Forchheimer_Inverse(InverseCase):
    def __init__(
        self,
        NumDomain=100,
        layer_size=[1] + [20] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.g = 1
        self.v = 1e-3
        self.e = 0.4
        self.v_e = dde.Variable(0.1)
        self.K = dde.Variable(0.1)
        self.v_e_true = 1e-3
        self.K_true = 1e-3
        super().__init__(
            "Inverse problem for Brinkman-Forchheimer model",
            external_trainable_variables=[self.v_e, self.K],
            NumDomain=NumDomain,
            use_output_transform=True,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            metrics=["l2 relative error"],
            visualization=Visualization_1D(x_label="t", y_label="y"),
        )

    def sol(self, x):
        H = 1
        r = (self.v * self.e / (1e-3 * 1e-3)) ** (0.5)
        return (
            self.g * 1e-3 / self.v * (1 - np.cosh(r * (x - H / 2)) / np.cosh(r * H / 2))
        )

    def gen_pde(self):
        def pde(x, y):
            du_xx = dde.grad.hessian(y, x)
            return -self.v_e / self.e * du_xx + self.v * y / self.K - self.g

        return pde

    def gen_geomtime(self):
        return dde.geometry.Interval(0, 1)

    def gen_data(self):
        num = 5
        xvals = np.linspace(1 / (num + 1), 1, num, endpoint=False)
        yvals = self.sol(xvals)
        ob_x, ob_u = np.reshape(xvals, (-1, 1)), np.reshape(yvals, (-1, 1))
        observe_u = dde.icbc.PointSetBC(ob_x, ob_u, component=0)
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            [observe_u],
            num_domain=self.NumDomain,
            num_boundary=0,
            num_test=500,
            solution=self.sol,
            train_distribution="uniform",
        )

    def output_transform(self, x, y):
        return x * (1 - x) * y

    def plot_result(self, solver, axes=None, exact=True):
        v_e_pred = bkd.to_numpy(self.v_e)
        K_pred = bkd.to_numpy(self.K)

        v_e_error = np.abs(self.v_e_true - v_e_pred)
        K_error = np.abs(self.K_true - K_pred)

        print(f"v_e true: {self.v_e_true}, K true: {self.K_true}")
        print(f"v_e pred: {v_e_pred}, K pred: {K_pred}")
        print(f"v_e error: {v_e_error}, K error: {K_error}")

        axes = self.Visualization.plot_line_1D(self, solver, exact, axes=axes)
        return axes


class Diffusion_Inverse(InverseCase):
    def __init__(
        self,
        NumDomain=40,
        layer_size=[2] + [32] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.C = dde.Variable(2.0)
        self.C_true = 1.0
        super().__init__(
            "Inverse problem for the diffusion equation",
            external_trainable_variables=self.C,
            NumDomain=NumDomain,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            metrics=["l2 relative error"],
            visualization=Visualization_2D(
                x_limit=[0, 1],
                y_limit=[-1, 1],
                x_label="t",
                y_label="x",
                feature_transform=lambda x: x[:, [1, 0]],
            ),
        )

    def sol(self, x):
        return np.sin(np.pi * x[:, 0:1]) * np.exp(-x[:, 1:])

    def gen_pde(self):
        def pde(x, y):
            dy_t = dde.grad.jacobian(y, x, i=0, j=1)
            dy_xx = dde.grad.hessian(y, x, i=0, j=0)
            return (
                dy_t
                - self.C * dy_xx
                + bkd.exp(-x[:, 1:])
                * (bkd.sin(np.pi * x[:, 0:1]) - np.pi**2 * bkd.sin(np.pi * x[:, 0:1]))
            )

        return pde

    def gen_geomtime(self):
        geom = dde.geometry.Interval(-1, 1)
        timedomain = dde.geometry.TimeDomain(0, 1)
        return dde.geometry.GeometryXTime(geom, timedomain)

    def gen_data(self):
        bc = dde.icbc.DirichletBC(
            self.geomtime, self.sol, lambda _, on_boundary: on_boundary
        )
        ic = dde.icbc.IC(self.geomtime, self.sol, lambda _, on_initial: on_initial)

        observe_x = np.vstack((np.linspace(-1, 1, num=10), np.full((10), 1))).T
        observe_y = dde.icbc.PointSetBC(observe_x, self.sol(observe_x), component=0)
        return dde.data.TimePDE(
            self.geomtime,
            self.pde,
            [bc, ic, observe_y],
            num_domain=self.NumDomain,
            num_boundary=20,
            num_initial=10,
            num_test=10000,
            anchors=observe_x,
            solution=self.sol,
        )

    def plot_result(self, solver, colorbar=[0, 0, 0]):
        C_pred = bkd.to_numpy(self.C)

        C_error = np.abs(self.C_true - C_pred)

        print(f"C true: {self.C_true}")
        print(f"C pred: {C_pred}")
        print(f"C error: {C_error}")

        X = np.array(
            [[x, t] for x in np.linspace(-1, 1, 1000) for t in np.linspace(0, 1, 1000)]
        )
        y = self.sol(X)
        model_y = solver.model.predict(X)

        fig, axes = self.Visualization.plot_exact_predict_error_2D(
            X, y, model_y, shape=[1000, 1000], title=solver.name, colorbar=colorbar
        )
        return fig, axes


class Diffusion_Reaction_Inverse(InverseCase):
    def __init__(
        self,
        NumDomain=2000,
        layer_size=[2] + [20] * 3 + [2],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.kf = dde.Variable(0.05)
        self.D = dde.Variable(1.0)

        self.kf_true = 2e-3
        self.D_true = 0.1
        super().__init__(
            "Inverse problem for the diffusion-reaction system",
            external_trainable_variables=[self.kf, self.D],
            NumDomain=NumDomain,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            visualization=Visualization_2D(
                x_limit=[0, 10],
                y_limit=[0, 1],
                x_label="t",
                y_label="x",
                feature_transform=lambda X: X[:, [1, 0]],
            ),
        )

    def gen_testdata(self):
        import os

        basepath = os.path.abspath(__file__)
        folder = os.path.dirname(os.path.dirname(basepath))
        data_path = os.path.join(folder, "data/reaction.npz")
        data = np.load(data_path)
        t, x, ca, cb = data["t"], data["x"], data["Ca"], data["Cb"]
        xx, tt = np.meshgrid(x, t)
        X = np.vstack((np.ravel(xx), np.ravel(tt))).T
        Ca = ca.flatten()[:, None]
        Cb = cb.flatten()[:, None]
        return X, np.hstack((Ca, Cb))

    def gen_pde(self):
        def pde(x, y):
            ca, cb = y[:, 0:1], y[:, 1:2]
            dca_t = dde.grad.jacobian(y, x, i=0, j=1)
            dca_xx = dde.grad.hessian(y, x, component=0, i=0, j=0)
            dcb_t = dde.grad.jacobian(y, x, i=1, j=1)
            dcb_xx = dde.grad.hessian(y, x, component=1, i=0, j=0)
            eq_a = dca_t - 1e-3 * self.D * dca_xx + self.kf * ca * cb**2
            eq_b = dcb_t - 1e-3 * self.D * dcb_xx + 2 * self.kf * ca * cb**2
            return [eq_a, eq_b]

        return pde

    def gen_geomtime(self):
        geom = dde.geometry.Interval(0, 1)
        timedomain = dde.geometry.TimeDomain(0, 10)
        return dde.geometry.GeometryXTime(geom, timedomain)

    def gen_data(self):
        def fun_bc(x):
            return 1 - x[:, 0:1]

        def fun_init(x):
            return np.exp(-20 * x[:, 0:1])

        bc_a = dde.icbc.DirichletBC(
            self.geomtime, fun_bc, lambda _, on_boundary: on_boundary, component=0
        )
        bc_b = dde.icbc.DirichletBC(
            self.geomtime, fun_bc, lambda _, on_boundary: on_boundary, component=1
        )
        ic1 = dde.icbc.IC(
            self.geomtime, fun_init, lambda _, on_initial: on_initial, component=0
        )
        ic2 = dde.icbc.IC(
            self.geomtime, fun_init, lambda _, on_initial: on_initial, component=1
        )

        observe_x, C = self.get_testdata()
        [Ca, Cb] = np.hsplit(C, 2)
        observe_y1 = dde.icbc.PointSetBC(observe_x, Ca, component=0)
        observe_y2 = dde.icbc.PointSetBC(observe_x, Cb, component=1)
        return dde.data.TimePDE(
            self.geomtime,
            self.pde,
            [bc_a, bc_b, ic1, ic2, observe_y1, observe_y2],
            num_domain=self.NumDomain,
            num_boundary=100,
            num_initial=100,
            num_test=5000,
            anchors=observe_x,
        )

    def plot_result(self, solver, colorbar=[0, 0, 0]):
        kf_pred = bkd.to_numpy(self.kf)
        D_pred = bkd.to_numpy(self.D)

        kf_error = np.abs(kf_pred - self.kf_true)
        D_error = np.abs(D_pred - self.D_true)

        print(f"kf true: {self.kf_true}, D true: {self.D_true}")
        print(f"kf pred: {kf_pred}, D pred: {D_pred}")
        print(f"kf error: {kf_error}, D error: {D_error}")

        X, y = self.get_testdata()
        model_y = solver.model.predict(X)

        fig0, axes0 = self.Visualization.plot_exact_predict_error_2D(
            X,
            y[:, 0],
            model_y[:, 0],
            shape=[201, 201],
            title=solver.name + " at dim 0",
            colorbar=colorbar,
        )

        fig1, axes1 = self.Visualization.plot_exact_predict_error_2D(
            X,
            y[:, 1],
            model_y[:, 1],
            shape=[201, 201],
            title=solver.name + " at dim 1",
            colorbar=colorbar,
        )
        return [fig0, axes0, fig1, axes1]


class Navier_Stokes_Incompressible_Flow_Around_Cylinder_Inverse(InverseCase):
    def __init__(
        self,
        NumDomain=700,
        layer_size=[3] + [50] * 6 + [3],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.C1 = dde.Variable(0.0)
        self.C2 = dde.Variable(0.0)

        self.C1_true = 1.0
        self.C2_true = 0.01
        super().__init__(
            "Inverse problem for the Navier-Stokes equation of incompressible flow around cylinder",
            external_trainable_variables=[self.C1, self.C2],
            NumDomain=NumDomain,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            visualization=Visualization_2D(
                x_limit=[1, 8], y_limit=[-2, 2], x_label="x1", y_label="x2"
            ),
        )

    def gen_testdata(self, num=7000):
        import os
        from scipy.io import loadmat

        basepath = os.path.abspath(__file__)
        folder = os.path.dirname(os.path.dirname(basepath))
        data_path = os.path.join(folder, "data/cylinder_nektar_wake.mat")
        data = loadmat(data_path)
        U_star = data["U_star"]  # N x 2 x T
        P_star = data["p_star"]  # N x T
        t_star = data["t"]  # T x 1
        X_star = data["X_star"]  # N x 2
        N = X_star.shape[0]
        T = t_star.shape[0]
        # Rearrange Data
        XX = np.tile(X_star[:, 0:1], (1, T))  # N x T
        YY = np.tile(X_star[:, 1:2], (1, T))  # N x T
        TT = np.tile(t_star, (1, N)).T  # N x T
        UU = U_star[:, 0, :]  # N x T
        VV = U_star[:, 1, :]  # N x T
        PP = P_star  # N x T
        x = XX.flatten()[:, None]  # NT x 1
        y = YY.flatten()[:, None]  # NT x 1
        t = TT.flatten()[:, None]  # NT x 1
        u = UU.flatten()[:, None]  # NT x 1
        v = VV.flatten()[:, None]  # NT x 1
        p = PP.flatten()[:, None]  # NT x 1
        # training domain: X × Y = [1, 8] × [−2, 2] and T = [0, 7]
        data1 = np.concatenate([x, y, t, u, v, p], 1)
        data2 = data1[:, :][data1[:, 2] <= 7]
        data3 = data2[:, :][data2[:, 0] >= 1]
        data4 = data3[:, :][data3[:, 0] <= 8]
        data5 = data4[:, :][data4[:, 1] >= -2]
        data_domain = data5[:, :][data5[:, 1] <= 2]
        # choose number of training points: num =7000
        idx = np.random.choice(data_domain.shape[0], num, replace=False)
        x_train = data_domain[idx, 0:1]
        y_train = data_domain[idx, 1:2]
        t_train = data_domain[idx, 2:3]
        u_train = data_domain[idx, 3:4]
        v_train = data_domain[idx, 4:5]
        p_train = data_domain[idx, 5:6]
        return [
            np.hstack((x_train, y_train, t_train)),
            np.hstack((u_train, v_train, p_train)),
        ]

    def gen_pde(self):
        def Navier_Stokes_Equation(x, y):
            u = y[:, 0:1]
            v = y[:, 1:2]
            p = y[:, 2:3]
            du_x = dde.grad.jacobian(y, x, i=0, j=0)
            du_y = dde.grad.jacobian(y, x, i=0, j=1)
            du_t = dde.grad.jacobian(y, x, i=0, j=2)
            dv_x = dde.grad.jacobian(y, x, i=1, j=0)
            dv_y = dde.grad.jacobian(y, x, i=1, j=1)
            dv_t = dde.grad.jacobian(y, x, i=1, j=2)
            dp_x = dde.grad.jacobian(y, x, i=2, j=0)
            dp_y = dde.grad.jacobian(y, x, i=2, j=1)
            du_xx = dde.grad.hessian(y, x, component=0, i=0, j=0)
            du_yy = dde.grad.hessian(y, x, component=0, i=1, j=1)
            dv_xx = dde.grad.hessian(y, x, component=1, i=0, j=0)
            dv_yy = dde.grad.hessian(y, x, component=1, i=1, j=1)
            continuity = du_x + dv_y
            x_momentum = (
                du_t
                + self.C1 * (u * du_x + v * du_y)
                + dp_x
                - self.C2 * (du_xx + du_yy)
            )
            y_momentum = (
                dv_t
                + self.C1 * (u * dv_x + v * dv_y)
                + dp_y
                - self.C2 * (dv_xx + dv_yy)
            )
            return [continuity, x_momentum, y_momentum]

        return Navier_Stokes_Equation

    def gen_geomtime(self):
        geom = dde.geometry.Rectangle([1.0, -2.0], [8.0, 2.0])
        timedomain = dde.geometry.TimeDomain(0, 7)
        return dde.geometry.GeometryXTime(geom, timedomain)

    def gen_data(self):
        X, y = self.gen_testdata(num=7000)
        ob_x, ob_y, ob_t = np.hsplit(X, 3)
        ob_u, ob_v, ob_p = np.hsplit(y, 3)
        ob_xyt = np.hstack((ob_x, ob_y, ob_t))
        observe_u = dde.icbc.PointSetBC(ob_xyt, ob_u, component=0)
        observe_v = dde.icbc.PointSetBC(ob_xyt, ob_v, component=1)
        return dde.data.TimePDE(
            self.geomtime,
            self.pde,
            [observe_u, observe_v],
            num_domain=self.NumDomain,
            num_boundary=200,
            num_initial=100,
            anchors=ob_xyt,
        )

    def plot_result(self, solver):
        C1_pred = bkd.to_numpy(self.C1)
        C2_pred = bkd.to_numpy(self.C2)

        C1_error = np.abs(C1_pred - self.C1_true)
        C2_error = np.abs(C2_pred - self.C2_true)

        print(f"C1 true: {self.C1_true}, C2 true: {self.C2_true}")
        print(f"C1 pred: {C1_pred}, C2 pred: {C2_pred}")
        print(f"C1 error: {C1_error}, C2 error: {C2_error}")

        import matplotlib.pyplot as plt
        import matplotlib.animation as animation

        fig, axes = plt.subplots()
        result_t = []
        for t in np.linspace(0, 7, 100):
            X = np.array(
                [
                    [x1, x2]
                    for x1 in np.linspace(1, 8, 1000)
                    for x2 in np.linspace(-2, 2, 1000)
                ]
            )
            X = np.hstack((X, np.full((X.shape[0], 1), t)))
            model_y = solver.model.predict(X)[:, 0]

            axs = self.Visualization.plot_heatmap_2D(
                X, model_y, shape=[1000, 1000], axes=axes, title=solver.name
            )
            result_t.append([axs])

        ani = animation.ArtistAnimation(fig, result_t)
        ani.save(
            "./result/Navier_Stokes_Incompressible_Flow_Around_Cylinder_Inverse.gif",
            writer="pillow",
        )



import numpy as np
import deepxde as dde
from ..cases.PDECases import AllenCahn
from ..solver.PDESolver import PINNSolver
import time

dde.optimizers.config.set_LBFGS_options(maxiter=1000)


class RAR_G(PINNSolver):
    def __init__(self, PDECase, iter=100):
        self.iter = iter
        super().__init__(name="RAR_G", PDECase=PDECase)

    def closure(self):
        self.train_step()

        for i in range(self.iter):
            X = self.PDECase.geomtime.random_points(100000)
            Y = np.abs(self.model.predict(X, operator=self.PDECase.pde))[:, 0]
            t1 = time.time()
            X_ids = np.argpartition(Y, self.PDECase.NumDomain // 100, axis=0)
            print(f"{time.time()-t1} s")
            self.PDECase.data.add_anchors(X[X_ids])

            self.train_step(iterations=1000)


if __name__ == "__main__":
    PDECase = AllenCahn(NumDomain=2000 // 2)
    solver = RAR_G(PDECase=PDECase)
    solver.train()
    solver.save(add_time=True)



from ..cases.PDECases import Burgers
from ..solver.PDESolver import PINNSolver


class PINN(PINNSolver):
    def __init__(self, PDECase):
        super().__init__(name="PINN", PDECase=PDECase)


if __name__ == "__main__":
    PDECase = Burgers(NumDomain=2000)
    solver = PINN(PDECase=PDECase)
    solver.train()
    solver.save(add_time=True)



import numpy as np
import deepxde as dde
import deepxde.backend as bkd
from .PDECases import PDECases
from abc import abstractmethod
from ..utils.Visualization import *


class PoissonCase1D(PDECases):
    def __init__(
        self,
        name,
        NumDomain=2000,
        Interval=None,
        use_output_transform=False,
        layer_size=[2] + [32] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
        metrics=["l2 relative error"],
        loss_weights=None,
        external_trainable_variables=None,
    ):
        self.Interval = Interval
        super().__init__(
            name=name,
            NumDomain=NumDomain,
            use_output_transform=use_output_transform,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            metrics=metrics,
            loss_weights=loss_weights,
            external_trainable_variables=external_trainable_variables,
            visualization=Visualization_1D(x_label="x", y_label="y"),
        )

    @abstractmethod
    def func(self, x):
        pass

    @abstractmethod
    def gen_data(self):
        pass

    def gen_geomtime(self):
        return dde.geometry.Interval(self.Interval[0], self.Interval[1])

    def gen_pde(self):
        def pde(x, y):
            dy_xx = dde.grad.hessian(y, x)
            return dy_xx - self.func(x)

        return pde

    def plot_result(self, solver, axes=None, exact=True):
        axes = self.Visualization.plot_line_1D(self, solver, exact, axes=axes)
        return axes


class PoissonCase2D(PDECases):
    def __init__(
        self,
        name,
        NumDomain=2000,
        x_min=None,
        x_max=None,
        use_output_transform=False,
        layer_size=[2] + [32] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
        metrics=["l2 relative error"],
        loss_weights=None,
        external_trainable_variables=None,
    ):
        self.x_min = x_min
        self.x_max = x_max
        super().__init__(
            name=name,
            NumDomain=NumDomain,
            use_output_transform=use_output_transform,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            metrics=metrics,
            loss_weights=loss_weights,
            external_trainable_variables=external_trainable_variables,
            visualization=Visualization_2D(
                x_limit=[self.x_min[0], self.x_max[0]],
                y_limit=[self.x_min[1], self.x_max[1]],
                x_label="x1",
                y_label="x2",
            ),
        )

    @abstractmethod
    def func(self, x):
        pass

    @abstractmethod
    def gen_data(self):
        pass

    def gen_geomtime(self):
        return dde.geometry.Rectangle(self.x_min, self.x_max)

    def gen_pde(self):
        def pde(x, y):
            dy_xx = dde.grad.hessian(y, x, i=0, j=0)
            dy_yy = dde.grad.hessian(y, x, i=1, j=1)
            return -dy_xx - dy_yy - self.func(x)

        return pde

    def plot_result(self, solver, colorbar=[0, 0, 0]):
        X = np.array(
            [
                [x1, x2]
                for x1 in np.linspace(self.x_min[0], self.x_max[0], 1000)
                for x2 in np.linspace(self.x_min[1], self.x_max[1], 1000)
            ]
        )
        y = self.sol(X)
        y[self.geomtime.inside(X) == 0] = np.nan
        model_y = solver.model.predict(X)
        model_y[self.geomtime.inside(X) == 0] = np.nan

        fig, axes = self.Visualization.plot_exact_predict_error_2D(
            X, y, model_y, shape=[1000, 1000], title=solver.name, colorbar=colorbar
        )
        return fig, axes


class Poisson_1D_Dirichlet(PoissonCase1D):
    def __init__(
        self,
        NumDomain=16,
        layer_size=[1] + [50] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.Interval = [-1, 1]
        super().__init__(
            name="Poisson equation in 1D with Dirichlet boundary conditions",
            NumDomain=NumDomain,
            Interval=self.Interval,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
        )

    def func(self, x):
        return -np.pi**2 * bkd.sin(np.pi * x)

    def sol(self, x):
        return np.sin(np.pi * x)

    def gen_data(self):
        bc = dde.icbc.DirichletBC(
            self.geomtime, self.sol, lambda _, on_boundary: on_boundary
        )
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            bc,
            self.NumDomain,
            2,
            solution=self.sol,
            num_test=100,
        )


class Poisson_1D_Dirichlet_Neumann(PoissonCase1D):
    def __init__(
        self,
        NumDomain=16,
        layer_size=[1] + [50] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.Interval = [-1, 1]
        super().__init__(
            name="Poisson equation in 1D with Dirichlet/Neumann boundary conditions",
            NumDomain=NumDomain,
            Interval=self.Interval,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
        )

    def func(self, x):
        return 2

    def sol(self, x):
        return (x + 1) ** 2

    def gen_data(self):
        def boundary_l(x, on_boundary):
            return on_boundary and np.isclose(x[0], -1)

        def boundary_r(x, on_boundary):
            return on_boundary and np.isclose(x[0], 1)

        bc_l = dde.icbc.DirichletBC(self.geomtime, self.sol, boundary_l)
        bc_r = dde.icbc.NeumannBC(self.geomtime, lambda X: 2 * (X + 1), boundary_r)
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            [bc_l, bc_r],
            self.NumDomain,
            2,
            solution=self.sol,
            num_test=100,
        )


class Poisson_1D_Dirichlet_Robin(PoissonCase1D):
    def __init__(
        self,
        NumDomain=16,
        layer_size=[1] + [50] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.Interval = [-1, 1]
        super().__init__(
            name="Poisson equation in 1D with Dirichlet/Robin boundary conditions",
            NumDomain=NumDomain,
            Interval=self.Interval,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
        )

    def func(self, x):
        return 2

    def sol(self, x):
        return (x + 1) ** 2

    def gen_data(self):
        def boundary_l(x, on_boundary):
            return on_boundary and np.isclose(x[0], -1)

        def boundary_r(x, on_boundary):
            return on_boundary and np.isclose(x[0], 1)

        bc_l = dde.icbc.DirichletBC(self.geomtime, self.sol, boundary_l)
        bc_r = dde.icbc.RobinBC(self.geomtime, lambda X, y: y, boundary_r)
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            [bc_l, bc_r],
            self.NumDomain,
            2,
            solution=self.sol,
            num_test=100,
        )


class Poisson_1D_Dirichlet_Periodic(PoissonCase1D):
    def __init__(
        self,
        NumDomain=16,
        layer_size=[1] + [50] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.Interval = [-1, 1]
        super().__init__(
            name="Poisson equation in 1D with Dirichlet/Periodic boundary conditions",
            NumDomain=NumDomain,
            Interval=self.Interval,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
        )

    def func(self, x):
        return -np.pi**2 * bkd.sin(np.pi * x)

    def sol(self, x):
        return np.sin(np.pi * x)

    def gen_data(self):
        def boundary_l(x, on_boundary):
            return on_boundary and np.isclose(x[0], -1)

        def boundary_r(x, on_boundary):
            return on_boundary and np.isclose(x[0], 1)

        bc_l = dde.icbc.DirichletBC(self.geomtime, self.sol, boundary_l)
        bc2 = dde.icbc.PeriodicBC(self.geomtime, 0, boundary_r)
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            [bc_l, bc2],
            self.NumDomain,
            2,
            solution=self.sol,
            num_test=100,
        )


class Poisson_1D_Dirichlet_PointSetOperator(PoissonCase1D):
    def __init__(
        self,
        NumDomain=16,
        layer_size=[1] + [50] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.Interval = [-1, 1]
        super().__init__(
            name="Poisson equation in 1D with Dirichlet/PointSetOperator boundary conditions",
            NumDomain=NumDomain,
            Interval=self.Interval,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
        )

    def func(self, x):
        return 2

    def sol(self, x):
        return (x + 1) ** 2

    def gen_data(self):
        def boundary_l(x, on_boundary):
            return on_boundary and np.isclose(x[0], -1)

        def dy_x(x, y, X):
            dy_x = dde.grad.jacobian(y, x)
            return dy_x

        def d_func(x):
            return 2 * (x + 1)

        bc_l = dde.icbc.DirichletBC(self.geomtime, self.sol, boundary_l)
        boundary_pts = self.geomtime.random_boundary_points(2)
        r_boundary_pts = boundary_pts[np.isclose(boundary_pts, 1)].reshape(-1, 1)
        bc_r = dde.icbc.PointSetOperatorBC(r_boundary_pts, d_func(r_boundary_pts), dy_x)
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            [bc_l, bc_r],
            self.NumDomain,
            2,
            solution=self.sol,
            num_test=100,
        )


class Poisson_1D_Hard_Boundary(PoissonCase1D):
    def __init__(
        self,
        NumDomain=64,
        layer_size=[1] + [50] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.Interval = [0, np.pi]
        super().__init__(
            name="Poisson equation in 1D with hard boundary conditions",
            NumDomain=NumDomain,
            Interval=self.Interval,
            use_output_transform=True,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
        )

    def func(self, x):
        summation = sum([i * bkd.sin(i * x) for i in range(1, 5)])
        return -summation - 8 * bkd.sin(8 * x)

    def sol(self, x):
        summation = sum([np.sin(i * x) / i for i in range(1, 5)])
        return x + summation + np.sin(8 * x) / 8

    def gen_data(self):
        return dde.data.PDE(
            self.geomtime, self.pde, [], self.NumDomain, solution=self.sol, num_test=400
        )

    def output_transform(self, x, y):
        return x * (np.pi - x) * y + x


class Poisson_1D_Fourier_Net(PoissonCase1D):
    def __init__(
        self,
        NumDomain=1280,
        layer_size=[1] + [100] * 3 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.Interval = [0, 1]
        self.A = 2
        self.B = 50
        super().__init__(
            name="Poisson equation in 1D with Multi-scale Fourier feature networks",
            NumDomain=NumDomain,
            Interval=self.Interval,
            use_output_transform=True,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
        )

    def gen_compile(
        self,
        metrics=None,
        loss_weights=None,
        external_trainable_variables=None,
    ):
        def compile(
            model,
            optimizer,
            lr=None,
            loss="MSE",
            decay=("inverse time", 2000, 0.9),
        ):
            model.compile(
                optimizer,
                lr,
                loss,
                metrics,
                decay,
                loss_weights,
                external_trainable_variables,
            )

        return compile

    def func(self, x):
        result = -((np.pi * self.A) ** 2) * bkd.sin(np.pi * self.A * x) - 0.1 * (
            np.pi * self.B
        ) ** 2 * bkd.sin(np.pi * self.B * x)
        return result

    def sol(self, x):
        return np.sin(np.pi * self.A * x) + 0.1 * np.sin(np.pi * self.B * x)

    def gen_net(self, layer_size, activation, initializer):
        return dde.nn.MsFFN(layer_size, activation, initializer, sigmas=[1, 10])

    def gen_data(self):
        bc = dde.icbc.DirichletBC(
            self.geomtime, self.sol, lambda _, on_boundary: on_boundary
        )
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            bc,
            self.NumDomain,
            train_distribution="pseudo",
            solution=self.sol,
            num_test=10000,
        )


class Poisson_2D_L_Shaped(PoissonCase2D):
    def __init__(
        self,
        NumDomain=1200,
        layer_size=[2] + [50] * 4 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        super().__init__(
            name="Poisson equation over L-shaped domain",
            NumDomain=NumDomain,
            x_min=[-1, -1],
            x_max=[1, 1],
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            metrics=None,
        )

    def gen_data(self):
        def boundary(_, on_boundary):
            return on_boundary

        bc = dde.icbc.DirichletBC(self.geomtime, lambda x: 0, boundary)
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            bc,
            num_domain=self.NumDomain,
            num_boundary=120,
            num_test=1500,
        )

    def gen_geomtime(self):
        return dde.geometry.Polygon(
            [[0, 0], [1, 0], [1, -1], [-1, -1], [-1, 1], [0, 1]]
        )

    def func(self, x):
        return bkd.from_numpy(np.ones((x.shape[0], 1)))

    def gen_testdata(self):
        X = np.array(
            [[x, t] for x in np.linspace(-1, 1, 1000) for t in np.linspace(-1, 1, 1000)]
        )
        y = np.linspace(-1, 1, 1000).T
        return X, y

    def plot_result(self, solver, colorbar=None):
        from matplotlib import pyplot as plt

        X = np.array(
            [[x, t] for x in np.linspace(-1, 1, 1000) for t in np.linspace(-1, 1, 1000)]
        )
        model_y = solver.model.predict(X)
        model_y[self.geomtime.inside(X) == 0] = np.nan

        fig, axes = plt.subplots()
        ax = self.Visualization.plot_heatmap_2D(
            X, model_y, shape=[1000, 1000], axes=axes, title=solver.name
        )
        if colorbar:
            fig.colorbar(ax, ax=axes)
        plt.show()
        return fig, axes


class Poisson_1D_Unknown_Forcing_Field_Inverse(PoissonCase1D):
    def __init__(
        self,
        NumDomain=16,
        layer_size=[1, [20, 20], [20, 20], [20, 20], 2],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        self.Interval = [-1, 1]
        super().__init__(
            name="Inverse problem for the Poisson equation with unknown forcing field",
            NumDomain=NumDomain,
            Interval=self.Interval,
            use_output_transform=False,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            loss_weights=[1, 100, 1000],
            metrics=None,
        )

    def gen_pde(self):
        def pde(x, y):
            u, q = y[:, 0:1], y[:, 1:2]
            du_xx = dde.grad.hessian(y, x, component=0, i=0, j=0)
            return -du_xx + q

        return pde

    def func(self, x):
        return -np.pi**2 * np.sin(np.pi * x)

    def sol(self, x):
        return np.sin(np.pi * x)

    def gen_data(self):
        bc = dde.icbc.DirichletBC(
            self.geomtime, self.sol, lambda _, on_boundary: on_boundary, component=0
        )
        num = 100
        ob_x = np.linspace(-1, 1, num).reshape(num, 1)
        ob_u = self.sol(ob_x)
        observe_u = dde.icbc.PointSetBC(ob_x, ob_u, component=0)
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            [bc, observe_u],
            self.NumDomain,
            2,
            anchors=ob_x,
            num_test=1000,
        )

    def gen_net(self, layer_size, activation, initializer):
        return dde.nn.PFNN(layer_size, activation, initializer)


class Poisson_1D_Fractional_Inverse(PoissonCase1D):
    def __init__(
        self,
        NumDomain=20,
        layer_size=[1] + [20] * 4 + [1],
        activation="tanh",
        initializer="Glorot normal",
    ):
        self.alpha = dde.Variable(1.5)
        self.alpha_true = 1.8
        self.Interval = [-1, 1]
        super().__init__(
            name="Inverse problem for the fractional Poisson equation in 1D",
            NumDomain=NumDomain,
            Interval=self.Interval,
            external_trainable_variables=self.alpha,
            use_output_transform=True,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            # loss_weights=[1, 100],
            metrics=None,
        )

    def gen_pde(self):
        from scipy.special import gamma

        def fpde(x, y, int_mat):
            """(D_{0+}^alpha + D_{1-}^alpha) u(x)"""
            if isinstance(int_mat, (list, tuple)) and len(int_mat) == 3:
                int_mat = bkd.sparse_tensor(*int_mat)
                lhs = bkd.sparse_dense_matmul(int_mat, y)
            else:
                lhs = bkd.matmul(int_mat, y)
            lhs /= 2 * bkd.cos(self.alpha * np.pi / 2)
            rhs = gamma(bkd.to_numpy(self.alpha) + 2) * x
            return lhs - rhs[: bkd.size(lhs)]

        return fpde

    def func(self, x):
        pass

    def sol(self, x):
        return x * (np.abs(1 - x**2)) ** (self.alpha_true / 2)

    def gen_data(self):
        observe_x = np.linspace(-1, 1, num=20)[:, None]
        observe_y = dde.icbc.PointSetBC(observe_x, self.sol(observe_x))

        return dde.data.FPDE(
            self.geomtime,
            self.pde,
            self.alpha,
            observe_y,
            [101],
            meshtype="static",
            anchors=observe_x,
            solution=self.sol,
        )
        # L-BFGS optimizer is not supported for dynamic meshtype.
        # return dde.data.FPDE(
        #    self.geomtime,
        #    self.pde,
        #    self.alpha,
        #    observe_y,
        #    [100],
        #    meshtype="dynamic",
        #    num_domain=self.NumDomain,
        #    anchors=observe_x,
        #    solution=self.sol,
        #    num_test=100,
        # )

    def output_transform(self, x, y):
        return (1 - x**2) * y

    def plot_result(self, solver):
        alpha_pred = bkd.to_numpy(self.alpha)

        alpha_error = np.abs(alpha_pred - self.alpha_true)

        print(f"alpha true: {self.alpha_true}")
        print(f"alpha pred: {alpha_pred}")
        print(f"alpha error: {alpha_error}")
        super().plot_result(solver)


class Poisson_2D_Fractional_Inverse(PDECases):
    def __init__(
        self,
        NumDomain=64,
        layer_size=[2] + [20] * 4 + [1],
        activation="tanh",
        initializer="Glorot normal",
    ):
        self.alpha = dde.Variable(1.5)
        self.alpha_true = 1.8
        self.Interval = [-1, 1]
        visualization = Visualization_2D(
            x_limit=[-1, 1], y_limit=[-1, 1], x_label="x1", y_label="x2"
        )
        super().__init__(
            name="Inverse problem for the fractional Poisson equation in 2D",
            NumDomain=NumDomain,
            external_trainable_variables=self.alpha,
            use_output_transform=True,
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            loss_weights=[1, 100],
            metrics=None,
            visualization=visualization,
        )

    def gen_pde(self):
        from scipy.special import gamma

        def fpde(x, y, int_mat):
            r"""\int_theta D_theta^alpha u(x)"""
            if isinstance(int_mat, (list, tuple)) and len(int_mat) == 3:
                int_mat = bkd.sparse_tensor(*int_mat)
                lhs = bkd.sparse_dense_matmul(int_mat, y)
            else:
                lhs = bkd.matmul(int_mat, y)
            lhs = lhs[:, 0]
            lhs *= -bkd.exp(
                bkd.lgamma((1 - self.alpha) / 2) + bkd.lgamma((2 + self.alpha) / 2)
            ) / (2 * np.pi**1.5)
            x = x[: bkd.size(lhs)]
            alpha = bkd.to_numpy(self.alpha)
            rhs = (
                2**alpha
                * gamma(2 + alpha / 2)
                * gamma(1 + alpha / 2)
                * (1 - (1 + alpha / 2) * bkd.sum(x**2, 1))
            )
            return lhs - rhs

        return fpde

    def gen_geomtime(self):
        return dde.geometry.Disk([0, 0], 1)

    def sol(self, x):
        return (1 - np.linalg.norm(x, axis=1, keepdims=True) ** 2) ** (
            1 + self.alpha_true / 2
        )

    def gen_data(self):
        observe_x = self.geomtime.random_points(30)
        observe_y = dde.icbc.PointSetBC(observe_x, self.sol(observe_x))

        # L-BFGS optimizer is not supported for dynamic meshtype.
        return dde.data.FPDE(
            self.geomtime,
            self.pde,
            self.alpha,
            observe_y,
            [8, 100],
            num_domain=self.NumDomain,
            anchors=observe_x,
            solution=self.sol,
        )

    def output_transform(self, x, y):
        return (1 - bkd.sum(x**2, 1, keepdims=True)) * y

    def plot_result(self, solver, colorbar=[0, 0, 0]):
        alpha_pred = bkd.to_numpy(self.alpha)

        alpha_error = np.abs(alpha_pred - self.alpha_true)

        print(f"alpha true: {self.alpha_true}")
        print(f"alpha pred: {alpha_pred}")
        print(f"alpha error: {alpha_error}")

        X = np.array(
            [
                [x1, x2]
                for x1 in np.linspace(-1, 1, 1000)
                for x2 in np.linspace(-1, 1, 1000)
            ]
        )
        y = self.sol(X)
        y[self.geomtime.inside(X) == 0] = np.nan
        model_y = solver.model.predict(X)
        model_y[self.geomtime.inside(X) == 0] = np.nan

        fig, axes = self.Visualization.plot_exact_predict_error_2D(
            X, y, model_y, shape=[1000, 1000], title=solver.name, colorbar=colorbar
        )
        return fig, axes


class Poisson_2D_Peak(PoissonCase2D):
    def __init__(
        self,
        NumDomain=2000,
        layer_size=[2] + [32] * 6 + [1],
        activation="tanh",
        initializer="Glorot uniform",
    ):
        super().__init__(
            name="Two-dimensional peak problem",
            NumDomain=NumDomain,
            x_min=[-1, -1],
            x_max=[1, 1],
            layer_size=layer_size,
            activation=activation,
            initializer=initializer,
            metrics=None,
        )

    def func(self, x):
        x, y = x[:, 0:1], x[:, 1:2]
        f_1 = 1000 * (x**2 - x + y**2 - y + 0.5)
        f_2 = -4000 * bkd.exp(-f_1)
        return -f_1 * f_2

    def sol(self, x):
        return np.exp(-1000 * ((x[:, 0:1] - 0.5) ** 2 + (x[:, 1:2] - 0.5) ** 2))

    def gen_data(self):
        bc = dde.icbc.DirichletBC(
            self.geomtime, self.sol, lambda _, on_boundary: on_boundary
        )
        return dde.data.PDE(
            self.geomtime,
            self.pde,
            bc,
            self.NumDomain,
            100,
            solution=self.sol,
            num_test=2000,
        )



import deepxde as dde
from ..cases.PDECases import Burgers
from ..solver.PDESolver import PINNSolver


class gPINN(PINNSolver):
    def __init__(self, PDECase):
        self.add_gradient_enhanced_res(PDECase)
        super().__init__(name="gPINN", PDECase=PDECase)

    def add_gradient_enhanced_res(self, PDECase):
        pde = PDECase.pde
        pde_num_args = dde.utils.get_num_args(pde)
        if PDECase.loss_weights is not None:
            icbc_len = len(PDECase.data.bcs)
            x_dim = PDECase.geomtime.random_points(1).shape[1]
            res_dim = len(PDECase.loss_weights) - icbc_len
            g_res_dim = res_dim * x_dim
            loss_weights = (
                PDECase.loss_weights[:res_dim]
                + [1] * g_res_dim
                + PDECase.loss_weights[res_dim:]
            )
            PDECase.set_loss_weights(loss_weights)

        if pde_num_args == 2:

            def g_pde(x, y):
                g_res = []
                res = pde(x, y)

                if isinstance(res, list):
                    g_res.extend(res)
                    res_dim = len(res)
                else:
                    g_res.append(res)
                    res_dim = 1
                    res = [res]

                x_dim = x.shape[1]
                for i in range(res_dim):
                    for j in range(x_dim):
                        res_ij = dde.grad.jacobian(res[i], x, j=j)
                        g_res.append(res_ij)

                return g_res

        elif pde_num_args == 3:

            def g_pde(x, y, ex):
                g_res = []
                res = pde(x, y, ex)

                if isinstance(res, list):
                    g_res.extend(res)
                    res_dim = len(res)
                else:
                    g_res.append(res)
                    res_dim = 1
                    res = [res]

                x_dim = x.shape[1]
                for i in range(res_dim):
                    for j in range(x_dim):
                        res_ij = dde.grad.jacobian(res[i], x, j=j)
                        g_res.append(res_ij)

                return g_res

        PDECase.set_pde(g_pde)


if __name__ == "__main__":
    PDECase = Burgers(NumDomain=2000)
    solver = gPINN(PDECase=PDECase)
    solver.train()
    solver.save(add_time=True)



import os
import time
import numpy as np
import deepxde as dde

dde.config.set_default_float("float64")


class PINNSolver:
    def __init__(self, name, PDECase):
        self.name = name
        self.PDECase = PDECase

        self.model = dde.Model(PDECase.data, PDECase.net)
        self.error = []
        self.losshistory = None
        self.train_state = None
        self.train_cost = None

    def save(self, path=None, add_time=False):
        if path is None:
            path = f"./models/{self.name}_{self.PDECase.name}"
        if add_time:
            path += f'_{time.strftime("%Y%m%d_%H%M%S", time.localtime())}/'
        else:
            path += "/"

        if not os.path.exists(path):
            os.makedirs(path)
        self.model.save(path)
        dde.saveplot(
            self.losshistory,
            self.train_state,
            issave=True,
            isplot=False,
            output_dir=path,
        )
        np.savetxt(f"{path}{self.name}_error.txt", self.error)

    def plot_loss_history(self, axes=None, train=False, use_time=False):
        import matplotlib.pyplot as plt

        if axes is None:
            fig, axes = plt.subplots()
        else:
            fig = axes.get_figure()
        loss_train = np.sum(self.losshistory.loss_train, axis=1)
        loss_test = np.sum(self.losshistory.loss_test, axis=1)

        if use_time:
            x = (
                np.array(self.losshistory.steps)
                / self.losshistory.steps[-1]
                * self.train_cost
            )
            axes.set_xlabel("Time(s)")
        else:
            x = self.losshistory.steps
            axes.set_xlabel("Steps")

        if train:
            axes.semilogy(x, loss_train, label="Train loss")
            axes.semilogy(x, loss_test, label="Test loss")
        else:
            axes.semilogy(x, loss_test, label=self.name)

        axes.set_title(self.name)
        axes.set_ylabel("Loss")
        axes.legend()
        return fig, axes

    def eval(self):
        X, y = self.PDECase.get_testdata()
        y_pred = self.model.predict(X)
        error = dde.metrics.l2_relative_error(y, y_pred)
        print("L2 relative error:", error)
        self.error.append(np.array(error))

    def train_step(self, lr=1e-3, iterations=15000, callbacks=None, eval=False):
        self.PDECase.compile(self.model, "adam", lr=lr)
        self.model.train(iterations=iterations, callbacks=callbacks)
        self.PDECase.compile(self.model, "L-BFGS")
        self.losshistory, self.train_state = self.model.train()
        if eval:
            self.eval()

    def closure(self):
        self.train_step()

    def train(self):
        t_start = time.time()
        print(f"Model {self.name} is training...")

        self.closure()

        t_end = time.time()
        self.train_cost = t_end - t_start



import numpy as np


def transform_uniform_to_normal_2D(X):
    max_X = np.max(X, axis=0)
    min_X = np.min(X, axis=0)
    X = (X - min_X) / (max_X - min_X)

    X_new = np.zeros_like(X)
    X_new[:, 0] = (-2 * np.log(X[:, 0])) ** (1.0 / 2) * np.cos(2 * np.pi * X[:, 1])
    X_new[:, 1] = (-2 * np.log(X[:, 0])) ** (1.0 / 2) * np.sin(2 * np.pi * X[:, 1])
    return X_new


def transform_normal_to_truncated_normal_on_geomtime(X, geomtime, mul=0, sigma=1):
    X = X * sigma + mul
    X = X[geomtime.inside(X)]
    return X



import numpy as np
import deepxde as dde
import deepxde.backend as bkd
from ..cases.PDECases import Burgers
from ..solver.PDESolver import PINNSolver

dde.optimizers.config.set_LBFGS_options(maxiter=1000)


class R3Sampling(PINNSolver):
    def __init__(
        self, PDECase, max_iter=100, causally_sampling=False, beta_lr=0.001, tol=20
    ):
        name = "R3Sampling"
        if causally_sampling:
            name += "_causal"
            self.set_causal_pde(PDECase)
        super().__init__(name=name, PDECase=PDECase)
        self.gamma = -0.5
        self.alpha = 5
        self.max_iter = max_iter
        self.causally_sampling = causally_sampling
        self.beta_lr = beta_lr
        self.tol = tol

    def set_causal_pde(self, PDECase):
        pde = PDECase.pde
        pde_num_args = dde.utils.get_num_args(pde)

        if pde_num_args == 2:

            def g_pde(x, y):
                res = pde(x, y)
                causal_weight = self.causal_weight(bkd.to_numpy(x), tensor=True)
                causal_weight = bkd.expand_dims(causal_weight, axis=1)

                if not isinstance(res, list):
                    res = [res]

                for res_i in res:
                    res_i = res_i * causal_weight

                return res

        elif pde_num_args == 3:

            def g_pde(x, y, ex):
                res = pde(x, y, ex)

                causal_weight = self.causal_weight(bkd.to_numpy(x), tensor=True)

                if not isinstance(res, list):
                    res = [res]

                for res_i in res:
                    res_i *= causal_weight

                return res

        PDECase.set_pde(g_pde)

    def update_gamma(self, loss):
        self.gamma += self.beta_lr * np.exp(-self.tol * loss)

    def causal_weight(self, X, tensor=False):
        X = X[:, -1]
        X = X / (max(X) - min(X))
        weight = (1 - np.tanh(self.alpha * (X - self.gamma))) / 2
        weight = weight.reshape(-1)
        if tensor:
            weight = bkd.pow(bkd.from_numpy(weight), 1 / 2)
        return weight

    def residual(self, X, causally_sampling=False):
        res = np.abs(self.model.predict(X, operator=self.PDECase.pde))
        res = np.sum(np.squeeze(res).T, axis=1)
        if causally_sampling:
            res = res * self.causal_weight(X)
            self.update_gamma(np.mean(res))
        return res

    def closure(self):
        self.train_step()

        for i in range(self.max_iter):
            X = self.PDECase.data.train_x_all
            res = self.residual(X, self.causally_sampling)
            threshold = np.mean(res)
            X_retained = X[np.where(res > threshold)]
            self.PDECase.data.replace_with_anchors(X_retained)

            N_r = len(X) - len(X_retained)
            if N_r > 0:
                X_resample = self.PDECase.geomtime.random_points(N_r)
                self.PDECase.data.add_anchors(X_resample)

            self.train_step(iterations=1000)


if __name__ == "__main__":
    PDECase = Burgers(NumDomain=2000)
    solver = R3Sampling(PDECase=PDECase)
    solver.train()
    solver.save(add_time=True)
